@@ -35,15 +35,17 @@
 typedef struct
 {
   IdeLspClient  *client;
-  gchar         *word;
-  gchar        **trigger_chars;
+  char          *word;
+  char         **trigger_chars;
+  char          *refilter_word;
+  guint          has_loaded : 1;
 } IdeLspCompletionProviderPrivate;
 
-static void provider_iface_init (IdeCompletionProviderInterface *iface);
+static void provider_iface_init (GtkSourceCompletionProviderInterface *iface);
 
 G_DEFINE_ABSTRACT_TYPE_WITH_CODE (IdeLspCompletionProvider, ide_lsp_completion_provider, IDE_TYPE_OBJECT,
                                   G_ADD_PRIVATE (IdeLspCompletionProvider)
-                                  G_IMPLEMENT_INTERFACE (IDE_TYPE_COMPLETION_PROVIDER, provider_iface_init))
+                                  G_IMPLEMENT_INTERFACE (GTK_SOURCE_TYPE_COMPLETION_PROVIDER, provider_iface_init))
 
 enum {
   PROP_0,
@@ -53,6 +55,11 @@ enum {
 
 static GParamSpec *properties [N_PROPS];
 
+static void
+ide_lsp_completion_provider_real_load (IdeLspCompletionProvider *self)
+{
+}
+
 static void
 ide_lsp_completion_provider_finalize (GObject *object)
 {
@@ -113,6 +120,8 @@ ide_lsp_completion_provider_class_init (IdeLspCompletionProviderClass *klass)
   object_class->get_property = ide_lsp_completion_provider_get_property;
   object_class->set_property = ide_lsp_completion_provider_set_property;
 
+  klass->load = ide_lsp_completion_provider_real_load;
+
   properties [PROP_CLIENT] =
     g_param_spec_object ("client",
                          "Client",
@@ -208,8 +217,8 @@ ide_lsp_completion_provider_set_client (IdeLspCompletionProvider *self,
 }
 
 static gint
-ide_lsp_completion_provider_get_priority (IdeCompletionProvider *provider,
-                                          IdeCompletionContext  *context)
+ide_lsp_completion_provider_get_priority (GtkSourceCompletionProvider *provider,
+                                          GtkSourceCompletionContext  *context)
 {
   return IDE_LSP_COMPLETION_PROVIDER_PRIORITY;
 }
@@ -251,20 +260,20 @@ ide_lsp_completion_provider_complete_cb (GObject      *object,
 }
 
 static void
-ide_lsp_completion_provider_populate_async (IdeCompletionProvider *provider,
-                                            IdeCompletionContext  *context,
-                                            GCancellable          *cancellable,
-                                            GAsyncReadyCallback    callback,
-                                            gpointer               user_data)
+ide_lsp_completion_provider_populate_async (GtkSourceCompletionProvider *provider,
+                                            GtkSourceCompletionContext  *context,
+                                            GCancellable                *cancellable,
+                                            GAsyncReadyCallback          callback,
+                                            gpointer                     user_data)
 {
   IdeLspCompletionProvider *self = (IdeLspCompletionProvider *)provider;
   IdeLspCompletionProviderPrivate *priv = ide_lsp_completion_provider_get_instance_private (self);
-  IdeCompletionActivation activation;
+  GtkSourceCompletionActivation activation;
   g_autoptr(IdeTask) task = NULL;
   g_autoptr(GVariant) params = NULL;
-  g_autofree gchar *uri = NULL;
+  g_autofree char *uri = NULL;
   GtkTextIter iter, end;
-  GtkTextBuffer *buffer;
+  GtkSourceBuffer *buffer;
   gint trigger_kind;
   gint line;
   gint column;
@@ -272,7 +281,16 @@ ide_lsp_completion_provider_populate_async (IdeCompletionProvider *provider,
   IDE_ENTRY;
 
   g_assert (IDE_IS_LSP_COMPLETION_PROVIDER (self));
-  g_assert (IDE_IS_COMPLETION_CONTEXT (context));
+  g_assert (GTK_SOURCE_IS_COMPLETION_CONTEXT (context));
+
+  if (!priv->has_loaded)
+    {
+      priv->has_loaded = TRUE;
+      IDE_LSP_COMPLETION_PROVIDER_GET_CLASS (self)->load (self);
+    }
+
+  g_clear_pointer (&priv->refilter_word, g_free);
+  g_clear_pointer (&priv->word, g_free);
 
   task = ide_task_new (self, cancellable, callback, user_data);
   ide_task_set_source_tag (task, ide_lsp_completion_provider_populate_async);
@@ -286,19 +304,18 @@ ide_lsp_completion_provider_populate_async (IdeCompletionProvider *provider,
       IDE_EXIT;
     }
 
-  ide_completion_context_get_bounds (context, &iter, &end);
+  gtk_source_completion_context_get_bounds (context, &iter, &end);
 
-  activation = ide_completion_context_get_activation (context);
+  activation = gtk_source_completion_context_get_activation (context);
 
-  if (activation == IDE_COMPLETION_TRIGGERED)
+  if (activation == GTK_SOURCE_COMPLETION_ACTIVATION_INTERACTIVE)
     trigger_kind = 2;
   else
     trigger_kind = 1;
 
-  g_clear_pointer (&priv->word, g_free);
-  priv->word = ide_completion_context_get_word (context);
+  priv->word = gtk_source_completion_context_get_word (context);
 
-  buffer = ide_completion_context_get_buffer (context);
+  buffer = gtk_source_completion_context_get_buffer (context);
   uri = ide_buffer_dup_uri (IDE_BUFFER (buffer));
 
   line = gtk_text_iter_get_line (&iter);
@@ -328,15 +345,15 @@ ide_lsp_completion_provider_populate_async (IdeCompletionProvider *provider,
 }
 
 static GListModel *
-ide_lsp_completion_provider_populate_finish (IdeCompletionProvider  *provider,
-                                             GAsyncResult           *result,
-                                             GError                **error)
+ide_lsp_completion_provider_populate_finish (GtkSourceCompletionProvider  *provider,
+                                             GAsyncResult                 *result,
+                                             GError                      **error)
 {
   GListModel *ret;
 
   IDE_ENTRY;
 
-  g_assert (IDE_IS_COMPLETION_PROVIDER (provider));
+  g_assert (GTK_SOURCE_IS_COMPLETION_PROVIDER (provider));
   g_assert (IDE_IS_TASK (result));
 
   ret = ide_task_propagate_object (IDE_TASK (result), error);
@@ -344,45 +361,47 @@ ide_lsp_completion_provider_populate_finish (IdeCompletionProvider  *provider,
   IDE_RETURN (ret);
 }
 
-static gboolean
-ide_lsp_completion_provider_refilter (IdeCompletionProvider *provider,
-                                      IdeCompletionContext  *context,
-                                      GListModel            *model)
+static void
+ide_lsp_completion_provider_refilter (GtkSourceCompletionProvider *provider,
+                                      GtkSourceCompletionContext  *context,
+                                      GListModel                  *model)
 {
+  IdeLspCompletionProvider *self = (IdeLspCompletionProvider *)provider;
+  IdeLspCompletionProviderPrivate *priv = ide_lsp_completion_provider_get_instance_private (self);
   IdeLspCompletionResults *results = (IdeLspCompletionResults *)model;
-  g_autofree gchar *word = NULL;
 
-  g_assert (IDE_IS_LSP_COMPLETION_PROVIDER (provider));
-  g_assert (IDE_IS_COMPLETION_CONTEXT (context));
+  g_assert (IDE_IS_LSP_COMPLETION_PROVIDER (self));
+  g_assert (GTK_SOURCE_IS_COMPLETION_CONTEXT (context));
   g_assert (IDE_IS_LSP_COMPLETION_RESULTS (results));
 
-  word = ide_completion_context_get_word (context);
-  ide_lsp_completion_results_refilter (results, word);
+  g_clear_pointer (&priv->refilter_word, g_free);
+  priv->refilter_word = gtk_source_completion_context_get_word (context);
 
-  return TRUE;
+  ide_lsp_completion_results_refilter (results, priv->refilter_word);
 }
 
 static void
-ide_lsp_completion_provider_display_proposal (IdeCompletionProvider   *provider,
-                                              IdeCompletionListBoxRow *row,
-                                              IdeCompletionContext    *context,
-                                              const gchar             *typed_text,
-                                              IdeCompletionProposal   *proposal)
+ide_lsp_completion_provider_display (GtkSourceCompletionProvider *provider,
+                                     GtkSourceCompletionContext  *context,
+                                     GtkSourceCompletionProposal *proposal,
+                                     GtkSourceCompletionCell     *cell)
 {
   IdeLspCompletionItem *item = (IdeLspCompletionItem *)proposal;
-  g_autofree gchar *markup = NULL;
+  IdeLspCompletionProvider *self = (IdeLspCompletionProvider *)provider;
+  IdeLspCompletionProviderPrivate *priv = ide_lsp_completion_provider_get_instance_private (self);
+  const char *typed_text;
 
   g_assert (IDE_IS_LSP_COMPLETION_PROVIDER (provider));
-  g_assert (IDE_IS_COMPLETION_LIST_BOX_ROW (row));
-  g_assert (IDE_IS_COMPLETION_CONTEXT (context));
+  g_assert (GTK_SOURCE_IS_COMPLETION_CONTEXT (context));
   g_assert (IDE_IS_LSP_COMPLETION_ITEM (proposal));
+  g_assert (GTK_SOURCE_IS_COMPLETION_CELL (cell));
 
-  markup = ide_lsp_completion_item_get_markup (item, typed_text);
+  if (priv->refilter_word)
+    typed_text = priv->refilter_word;
+  else
+    typed_text = priv->word;
 
-  ide_completion_list_box_row_set_icon_name (row, ide_lsp_completion_item_get_icon_name (item));
-  ide_completion_list_box_row_set_left (row, NULL);
-  ide_completion_list_box_row_set_center_markup (row, markup);
-  ide_completion_list_box_row_set_right (row, NULL);
+  ide_lsp_completion_item_display (item, cell, typed_text);
 }
 
 static void
@@ -406,35 +425,40 @@ ide_lsp_completion_provider_apply_additional_edits_cb (GObject      *object,
 }
 
 static void
-ide_lsp_completion_provider_activate_proposal (IdeCompletionProvider *provider,
-                                               IdeCompletionContext  *completion_context,
-                                               IdeCompletionProposal *proposal,
-                                               const GdkEventKey     *key)
+ide_lsp_completion_provider_activate (GtkSourceCompletionProvider *provider,
+                                      GtkSourceCompletionContext  *context,
+                                      GtkSourceCompletionProposal *proposal)
 {
-  g_autoptr(IdeSnippet) snippet = NULL;
-  GtkTextBuffer *buffer;
-  GtkTextView *view;
+  g_autoptr(GPtrArray) additional_text_edits = NULL;
+  g_autoptr(GtkSourceSnippet) snippet = NULL;
+  GtkSourceSnippetChunk *chunk;
+  GtkSourceBuffer *buffer;
+  GtkSourceView *view;
   GtkTextIter begin, end;
-  g_autoptr(GPtrArray) additional_text_edits;
-  IdeBufferManager *buffer_manager;
+  const char *text = NULL;
 
-  g_assert (IDE_IS_COMPLETION_PROVIDER (provider));
-  g_assert (IDE_IS_COMPLETION_CONTEXT (completion_context));
+  g_assert (GTK_SOURCE_IS_COMPLETION_PROVIDER (provider));
+  g_assert (GTK_SOURCE_IS_COMPLETION_CONTEXT (context));
   g_assert (IDE_IS_LSP_COMPLETION_ITEM (proposal));
 
-  buffer = ide_completion_context_get_buffer (completion_context);
-
+  buffer = gtk_source_completion_context_get_buffer (context);
   g_assert (IDE_IS_BUFFER (buffer));
 
-  view = ide_completion_context_get_view (completion_context);
+  view = gtk_source_completion_context_get_view (context);
+  g_assert (IDE_IS_SOURCE_VIEW (view));
 
   snippet = ide_lsp_completion_item_get_snippet (IDE_LSP_COMPLETION_ITEM (proposal));
+  if ((chunk = gtk_source_snippet_get_nth_chunk (snippet, 0)))
+    text = gtk_source_snippet_chunk_get_text (chunk);
 
-  gtk_text_buffer_begin_user_action (buffer);
-  if (ide_completion_context_get_bounds (completion_context, &begin, &end))
-    gtk_text_buffer_delete (buffer, &begin, &end);
-  ide_source_view_push_snippet (IDE_SOURCE_VIEW (view), snippet, &begin);
-  gtk_text_buffer_end_user_action (buffer);
+  gtk_text_buffer_begin_user_action (GTK_TEXT_BUFFER (buffer));
+  if (gtk_source_completion_context_get_bounds (context, &begin, &end))
+    {
+      gtk_text_buffer_delete (GTK_TEXT_BUFFER (buffer), &begin, &end);
+      ide_text_util_remove_common_prefix (&begin, text);
+    }
+  gtk_source_view_push_snippet (GTK_SOURCE_VIEW (view), snippet, &begin);
+  gtk_text_buffer_end_user_action (GTK_TEXT_BUFFER (buffer));
 
   additional_text_edits =
     ide_lsp_completion_item_get_additional_text_edits (IDE_LSP_COMPLETION_ITEM (proposal),
@@ -442,69 +466,45 @@ ide_lsp_completion_provider_activate_proposal (IdeCompletionProvider *provider,
   IDE_PTR_ARRAY_SET_FREE_FUNC (additional_text_edits, g_object_unref);
 
   if (additional_text_edits != NULL)
-    {
-      IdeContext *context = ide_object_get_context (IDE_OBJECT (provider));
-
-      g_return_if_fail (context != NULL);
-
-      buffer_manager = ide_buffer_manager_from_context (context);
-
-      ide_buffer_manager_apply_edits_async (buffer_manager,
-                                            IDE_PTR_ARRAY_STEAL_FULL (&additional_text_edits),
-                                            NULL,
-                                            ide_lsp_completion_provider_apply_additional_edits_cb,
-                                            NULL);
-    }
-}
-
-static gchar *
-ide_lsp_completion_provider_get_comment (IdeCompletionProvider *provider,
-                                         IdeCompletionProposal *proposal)
-{
-  g_assert (IDE_IS_COMPLETION_PROVIDER (provider));
-  g_assert (IDE_IS_LSP_COMPLETION_ITEM (proposal));
-
-  return g_strdup (ide_lsp_completion_item_get_detail (IDE_LSP_COMPLETION_ITEM (proposal)));
+    ide_buffer_manager_apply_edits_async (ide_buffer_manager_from_context (ide_object_get_context (IDE_OBJECT (provider))),
+                                          IDE_PTR_ARRAY_STEAL_FULL (&additional_text_edits),
+                                          NULL,
+                                          ide_lsp_completion_provider_apply_additional_edits_cb,
+                                          NULL);
 }
 
 static gboolean
-ide_lsp_completion_provider_is_trigger (IdeCompletionProvider *provider,
-                                        const GtkTextIter     *iter,
-                                        gunichar               ch)
+ide_lsp_completion_provider_is_trigger (GtkSourceCompletionProvider *provider,
+                                        const GtkTextIter           *iter,
+                                        gunichar                     ch)
 {
+  static const char * const default_trigger_chars[] = IDE_STRV_INIT (".");
   IdeLspCompletionProvider *self = (IdeLspCompletionProvider *)provider;
   IdeLspCompletionProviderPrivate *priv = ide_lsp_completion_provider_get_instance_private (self);
+  const char * const *trigger_chars;
 
   g_assert (IDE_IS_LSP_COMPLETION_PROVIDER (self));
   g_assert (iter != NULL);
 
-  if (priv->trigger_chars != NULL)
-    {
-      for (guint i = 0; priv->trigger_chars[i]; i++)
-        {
-          const gchar *trigger = priv->trigger_chars[i];
+  trigger_chars = priv->trigger_chars ? (const char * const *)priv->trigger_chars : default_trigger_chars;
 
-          /* Technically, since these are strings they can be more than
-           * one character long. But I haven't seen anything actually
-           * do that in the wild yet.
-           */
-          if (ch == g_utf8_get_char (trigger))
-            return TRUE;
-        }
+  for (guint i = 0; trigger_chars[i]; i++)
+    {
+      if (ch == g_utf8_get_char (trigger_chars[i]))
+        return TRUE;
     }
 
   return FALSE;
 }
 
 static void
-provider_iface_init (IdeCompletionProviderInterface *iface)
+provider_iface_init (GtkSourceCompletionProviderInterface *iface)
 {
   iface->get_priority = ide_lsp_completion_provider_get_priority;
   iface->populate_async = ide_lsp_completion_provider_populate_async;
   iface->populate_finish = ide_lsp_completion_provider_populate_finish;
   iface->refilter = ide_lsp_completion_provider_refilter;
-  iface->display_proposal = ide_lsp_completion_provider_display_proposal;
-  iface->activate_proposal = ide_lsp_completion_provider_activate_proposal;
-  iface->get_comment = ide_lsp_completion_provider_get_comment;
+  iface->display = ide_lsp_completion_provider_display;
+  iface->activate = ide_lsp_completion_provider_activate;
   iface->is_trigger = ide_lsp_completion_provider_is_trigger;
 }
